"""Read various spatial vector formats into a Spark DataFrame.

Basic usage
===========

Read the first layer from a file or files into a single Spark DataFrame:

.. code-block:: python

    sdf = read_vector_files(
        path="/path/to/files/",
        suffix=".ext",
    )

Filename pattern matching
=========================

Read files that begin with "abc" into a single Spark DataFrame:

.. code-block:: python

    sdf = read_vector_files(
        path="/path/to/files/",
        pattern="abc*",
        suffix=".ext",
    )

Read files that end with four digits into a single Spark DataFrame:

.. code-block:: python

    sdf = read_vector_files(
        path="/path/to/files/",
        pattern="*[0-9][0-9][0-9][0-9]",
        suffix=".ext",
    )

For more information on pattern matching using Unix shell-style wildcards, see
Python's `fnmatch`_ module.

Reading files from nested folders
=================================

By default, the library will only look within the specified folder. To enable
recursive searching of subdirectories, use the `recursive` argument.


Given the following folder structure::

    /path/to/files
    |    file_0.ext
    |    file_1.ext
    |
    |-- subfolder
    |        file_2.ext
    |        file_3.ext

.. code-block:: python

    sdf = read_vector_files(
        path="/path/to/files/",
        suffix=".ext",
    )

will read `file_0.ext` and `file_1.ext`, while

.. code-block:: python

    sdf = read_vector_files(
        path="/path/to/files/",
        suffix=".ext",
        recursive=True,
    )

will read `file_0.ext`, `file_1.ext`, `subfolder/file_2.ext`, and
`subfolder/file_3.ext`.

Reading layers
==============

Read a specific layer for a file or files, using layer name:

.. code-block:: python

    sdf = read_vector_files(
        path="/path/to/files/",
        suffix=".ext",
        layer_identifier="layer_name"
    )

or layer index:

.. code-block:: python

    sdf = read_vector_files(
        path="/path/to/files/",
        suffix=".ext",
        layer_identifier=1
    )

GDAL Virtual File Systems
=========================

Read compressed files using GDAL Virtual File Systems:

.. code-block:: python

    sdf = read_vector_files(
        path="/path/to/files/",
        suffix=".gz",
        layer_identifier="layer_name",
        vsi_prefix="/vsigzip/",
    )

For more information, see `GDAL Virtual File Systems`_.

User-defined Schema
===================

By default, a schema will be generated from the first file in the folder. For a
single tabular dataset that has been partitioned across several files, this will
work fine.

However, it won't work for a list format like GML, as not every file will contain
the same fields. In this case, you can define a schema yourself. You will also need
to set the coerce_to_schema flag to True.

.. code-block:: python

    schema = StructType(
        [
            StructField("id", LongType()),
            StructField("category", StringType()),
            StructField("geometry", BinaryType()),
        ]
    )

    sdf = read_vector_files(
        path="/path/to/files/",
        suffix=".ext",
        layer_identifier="layer_name",
        schema=schema,
        coerce_to_schema=True,
    )

Concurrency Strategy
====================

By default, the function will parallelise across files.

This should work well for single dataset that has been partitioned across several
files. Especially if it has been partition so that those individual files can be
comfortably read into memory on a single machine.

However, the function also provides a way of parallelising across chunks of rows
within a file or files.

.. code-block:: python

    sdf = read_vector_files(
        path="/path/to/files/",
        suffix=".ext",
        concurrency_strategy="rows",
    )

By default, a chunk will consist of 1 million rows but you can change this using the
`ideal_chunk_size` parameter.

.. code-block:: python

    sdf = read_vector_files(
        path="/path/to/files/",
        suffix=".ext",
        concurrency_strategy="rows",
        ideal_chunk_size=5_000_000,
    )

.. warning::
    Reading chunks adds a substantial overhead as files have to be opened to get a row
    count. The "rows" strategy should only be used for a single large file or a small
    number of large files.

.. _`fnmatch`: https://docs.python.org/3/library/fnmatch.html#module-fnmatch

.. _`GDAL Virtual File Systems`: https://gdal.org/user/virtual_file_systems.html

"""
try:
    from importlib.metadata import PackageNotFoundError, version  # type: ignore
except ImportError:  # pragma: no cover
    from importlib_metadata import PackageNotFoundError, version  # type: ignore


try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"

from contextlib import contextmanager
from types import MappingProxyType
from typing import Iterator, Optional, Tuple, Union

from numpy import bool_, float32, float64, int16, int32, int64, object_
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql import SparkSession
from pyspark.sql.types import (
    ArrayType,
    BinaryType,
    BooleanType,
    DoubleType,
    FloatType,
    IntegerType,
    LongType,
    ShortType,
    StringType,
    StructType,
)

from pyspark_vector_files._files import (
    _add_vsi_prefix,
    _create_chunks_sdf,
    _create_paths_sdf,
    _get_data_sources,
    _get_feature_counts,
    _get_layer_names,
    _get_paths,
    _get_sequence_of_chunks,
    _get_total_chunks,
)
from pyspark_vector_files._parallel_reader import (
    _generate_parallel_reader_for_chunks,
    _generate_parallel_reader_for_files,
)
from pyspark_vector_files._schema import (
    _create_schema_for_chunks,
    _create_schema_for_files,
)
from pyspark_vector_files._types import ConcurrencyStrategy

# See:
# https://gdal.org/python/index.html
# https://spark.apache.org/docs/latest/sql-ref-datatypes.html
# https://pandas.pydata.org/docs/user_guide/basics.html#dtypes
# https://numpy.org/doc/stable/reference/arrays.dtypes.html

OGR_TO_SPARK = MappingProxyType(
    {
        (0, 0): IntegerType(),  # OFTInteger
        (0, 1): BooleanType(),  # OFSTBoolean
        (0, 2): ShortType(),  # OFSTInt16
        (1, 0): ArrayType(IntegerType()),  # OFTIntegerList
        (1, 1): ArrayType(ShortType()),  # List of OFSTInt16
        (1, 2): ArrayType(BooleanType()),  # List of OFSTBoolean
        (2, 0): DoubleType(),  # OFTReal
        (2, 3): FloatType(),  # OFSTFloat32
        (3, 0): ArrayType(DoubleType()),  # OFTRealList
        (3, 3): ArrayType(FloatType()),  # List of OFSTFloat32
        (4, 0): StringType(),  # OFTString
        # ? Could OFSTJSON be MapType?
        (4, 4): StringType(),  # OFSTJSON
        (4, 5): StringType(),  # OFSTUUID
        (5, 0): ArrayType(StringType()),  # OFTStringList
        (6, 0): StringType(),  # OFTWideString
        (7, 0): ArrayType(StringType()),  # OFTWideStringList
        (8, 0): BinaryType(),  # OFTBinary
        (9, 0): StringType(),  # OFTDate
        (10, 0): StringType(),  # OFTTime
        (11, 0): StringType(),  # OFTDateTime
        (12, 0): LongType(),  # OFTInteger64
        (13, 0): ArrayType(LongType()),  # OFTInteger64List
    }
)

SPARK_TO_PANDAS = MappingProxyType(
    {
        IntegerType(): int32,
        BooleanType(): bool_,
        ShortType(): int16,
        ArrayType(IntegerType()): object_,
        ArrayType(BooleanType()): object_,
        ArrayType(ShortType()): object_,
        DoubleType(): float64,
        FloatType(): float32,
        ArrayType(DoubleType()): object_,
        ArrayType(FloatType()): object_,
        StringType(): object_,
        ArrayType(StringType()): object_,
        BinaryType(): object_,
        LongType(): int64,
        ArrayType(LongType()): object_,
    }
)


@contextmanager
def temporary_spark_context(
    configuration_key: str,
    new_configuration_value: str,
    spark: SparkSession = None,
) -> Iterator[SparkSession]:
    """Changes then resets spark configuration."""
    spark = spark if spark else SparkSession.getActiveSession()
    old_configuration_value = spark.conf.get(configuration_key)
    spark.conf.set(configuration_key, new_configuration_value)
    try:
        yield spark
    finally:
        spark.conf.set(configuration_key, old_configuration_value)


def read_vector_files(
    path: str,
    ogr_to_spark_type_map: MappingProxyType = OGR_TO_SPARK,
    pattern: str = "*",
    suffix: str = "*",
    recursive: bool = False,
    ideal_chunk_size: int = 1_000_000,
    geom_field_name: str = "geometry",
    geom_field_type: Tuple[int, int] = (8, 0),
    coerce_to_schema: bool = False,
    spark_to_pandas_type_map: MappingProxyType = SPARK_TO_PANDAS,
    concurrency_strategy: str = "files",
    vsi_prefix: Optional[str] = None,
    schema: Optional[StructType] = None,
    layer_identifier: Optional[Union[str, int]] = None,
) -> SparkDataFrame:
    """Read vector file(s) into a Spark DataFrame.

    Args:
        path (str): Path to a folder of vector files.
        ogr_to_spark_type_map (MappingProxyType): A mapping of OGR to Spark data
            types. Defaults to OGR_TO_SPARK.
        pattern (str): A filename pattern. This will be passed to to `pathlib`'s
            `Path.glob` method. For more information, see
            https://docs.python.org/3/library/pathlib.html#pathlib.Path.glob.
            Defaults to "*".
        suffix (str): A file extension pattern. This will be passed to to `pathlib`'s
            `Path.glob` method. For more information, see
            https://docs.python.org/3/library/pathlib.html#pathlib.Path.glob.
            Defaults to "*".
        recursive (bool): If True, recursive globbing in enabled. For more
            information, see
            https://docs.python.org/3/library/pathlib.html#pathlib.Path.rglob.
            Defaults to False.
        ideal_chunk_size (int): The max number of rows to be read into a chunk.
            Defaults to 1_000_000.
        geom_field_name (str): The name of the geometry column. Defaults to
            "geometry".
        geom_field_type (Tuple[int, int]): The data type of the geometry column when
            it is passed to Spark. Defaults to `(8, 0)`, which represents binary
            data.
        coerce_to_schema (bool): If True, all files or chunks will be forced to
            fit the supplied schema. Missing columns will be added and additional
            columns will be removed. Defaults to False.
        spark_to_pandas_type_map (MappingProxyType): A mapping of Spark to Pandas data
            types. Defaults to SPARK_TO_PANDAS.
        concurrency_strategy (str): The concurrency strategy to use, can be "files"
            or "rows". Defaults to "files".
        vsi_prefix (str, optional): The GDAL virtual file system prefix(es) to use.
            For more information, see https://gdal.org/user/virtual_file_systems.html.
            Defaults to None.
        schema (StructType, optional): A user-defined Spark schema. Defaults to None.
        layer_identifier (Union[str, int], optional): A layer name or index. If
            None is given, the first layer will be return. Defaults to None.

    Returns:
        SparkDataFrame: A Spark DataFrame with geometry encoded as
            `Well Known Binary (WKB)`_.

    .. _`Well Known Binary (WKB)`:
        https://libgeos.org/specifications/wkb/
    """
    _concurrency_strategy = ConcurrencyStrategy(concurrency_strategy)

    paths = _get_paths(
        path=path,
        pattern=pattern,
        suffix=suffix,
        recursive=recursive,
    )

    if vsi_prefix:
        paths = _add_vsi_prefix(
            paths=paths,
            vsi_prefix=vsi_prefix,
        )

    if _concurrency_strategy == ConcurrencyStrategy.FILES:

        number_of_partitions = len(paths)

        with temporary_spark_context(
            configuration_key="spark.sql.shuffle.partitions",
            new_configuration_value=str(number_of_partitions),
        ) as spark:

            df = _create_paths_sdf(
                spark=spark,
                paths=paths,
            )

            _schema = (
                schema
                if schema
                else _create_schema_for_files(
                    path=paths[0],
                    layer_identifier=layer_identifier,
                    ogr_to_spark_type_map=ogr_to_spark_type_map,
                    geom_field_name=geom_field_name,
                    geom_field_type=geom_field_type,
                )
            )

            parallel_read = _generate_parallel_reader_for_files(
                layer_identifier=layer_identifier,
                geom_field_name=geom_field_name,
                coerce_to_schema=coerce_to_schema,
                spark_to_pandas_type_map=spark_to_pandas_type_map,
                schema=_schema,
            )

            return (
                df.repartition(number_of_partitions, "path")
                .groupby("path")
                .applyInPandas(parallel_read, _schema)
            )

    else:
        data_sources = _get_data_sources(paths)

        layer_names = _get_layer_names(
            data_sources=data_sources,
            layer_identifier=layer_identifier,
        )

        feature_counts = _get_feature_counts(
            data_sources=data_sources,
            layer_names=layer_names,
        )

        sequence_of_chunks = _get_sequence_of_chunks(
            feature_counts=feature_counts,
            ideal_chunk_size=ideal_chunk_size,
        )

        number_of_partitions = _get_total_chunks(sequence_of_chunks)

        with temporary_spark_context(
            configuration_key="spark.sql.shuffle.partitions",
            new_configuration_value=str(number_of_partitions),
        ) as spark:

            df = _create_chunks_sdf(
                spark=spark,
                paths=paths,
                layer_names=layer_names,
                sequence_of_chunks=sequence_of_chunks,
            )

            _schema = (
                schema
                if schema
                else _create_schema_for_chunks(
                    data_source=data_sources[0],
                    layer_name=layer_names[0],
                    ogr_to_spark_type_map=ogr_to_spark_type_map,
                    geom_field_name=geom_field_name,
                    geom_field_type=geom_field_type,
                )
            )

            parallel_read = _generate_parallel_reader_for_chunks(
                geom_field_name=geom_field_name,
                coerce_to_schema=coerce_to_schema,
                spark_to_pandas_type_map=spark_to_pandas_type_map,
                schema=_schema,
            )

            return (
                df.repartition(number_of_partitions, "id")
                .groupby("id")
                .applyInPandas(parallel_read, _schema)
            )



from functools import singledispatch
from pathlib import Path
from typing import Optional, Tuple, Union

from more_itertools import pairwise
from osgeo.ogr import DataSource, Layer, Open
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql import Row, SparkSession
from pyspark.sql.functions import col, explode, monotonically_increasing_id
from pyspark.sql.types import (
    ArrayType,
    IntegerType,
    StringType,
    StructField,
    StructType,
)

from pyspark_vector_files._types import Chunks


def _get_paths(
    path: str, pattern: str, suffix: str, recursive: bool
) -> Tuple[str, ...]:
    """Returns full paths for all files in a path, with the given suffix."""
    _path = Path(path)

    if not _path.is_dir():
        raise NotADirectoryError(
            f'"{str(_path)}" is not a directory.',
        )

    _suffix = suffix.strip(".")

    if recursive:
        paths = tuple(str(path) for path in _path.rglob(f"{pattern}.{_suffix}"))
    else:
        paths = tuple(str(path) for path in _path.glob(f"{pattern}.{_suffix}"))

    if len(paths) == 0:
        raise FileNotFoundError(
            f'No files found for: path: "{path}", pattern: "{pattern}", suffix: "{suffix}", recursive: {recursive}.',  # noqa: B950
        )
    else:
        return paths


def _add_vsi_prefix(paths: Tuple[str, ...], vsi_prefix: str) -> Tuple[str, ...]:
    """Adds GDAL virtual file system prefix to the paths."""
    _vsi_prefix = vsi_prefix.strip("/")
    _paths = tuple(path.lstrip("/") for path in paths)
    return tuple("/" + _vsi_prefix + "//" + path for path in _paths)


def _get_data_sources(paths: Tuple[str, ...]) -> Tuple[DataSource, ...]:
    return tuple(Open(path) for path in paths)


def _get_data_source_layer_names(data_source: DataSource) -> Tuple[str, ...]:
    """Given a OGR DataSource, returns a sequence of layers."""
    return tuple(
        data_source.GetLayer(index).GetName()
        for index in range(data_source.GetLayerCount())
    )


@singledispatch
def _get_layer_name(
    layer_identifier: Optional[Union[str, int]],
    data_source: DataSource,
) -> Layer:
    """Returns the layer name, the layer name at an index, or the first layer name."""
    pass


@_get_layer_name.register
def _get_layer_name_none(
    layer_identifier: None,
    data_source: DataSource,
) -> Layer:
    """Returns the name of the first layer."""
    data_source_layer_names = _get_data_source_layer_names(
        data_source,
    )

    layer_name = data_source_layer_names[0]

    return layer_name


@_get_layer_name.register
def _get_layer_name_str(
    layer_identifier: str,
    data_source: DataSource,
) -> Layer:
    """Returns the given layer name, if that layer name exists."""
    data_source_layer_names = _get_data_source_layer_names(
        data_source,
    )

    if layer_identifier not in data_source_layer_names:
        raise ValueError(
            f"Expecting one of {data_source_layer_names} but received {layer_identifier}.",  # noqa B950
        )
    else:
        layer_name = layer_identifier

    return layer_name


@_get_layer_name.register
def _get_layer_name_int(
    layer_identifier: int,
    data_source: DataSource,
) -> Layer:
    """Returns the layer name at given index, if that index is valid."""
    data_source_layer_names = _get_data_source_layer_names(
        data_source,
    )

    number_of_layers = len(data_source_layer_names)

    if layer_identifier >= number_of_layers:
        raise ValueError(
            f"Expecting index between 0 and {number_of_layers} but received {layer_identifier}.",  # noqa B950
        )
    else:
        layer_name = data_source_layer_names[layer_identifier]

    return layer_name


def _get_layer_names(
    layer_identifier: Optional[Union[str, int]],
    data_sources: Tuple[DataSource, ...],
) -> Tuple[str, ...]:
    return tuple(
        _get_layer_name(layer_identifier, data_source=data_source)
        for data_source in data_sources
    )


def _get_feature_count(data_source: DataSource, layer_name: str) -> int:
    layer = data_source.GetLayer(layer_name)
    # ! Explicitly coerce `feat_count` to `int` to prevent `mypy` `[no-any-return]`
    feat_count: int = layer.GetFeatureCount()
    if feat_count == -1:
        feat_count = layer.GetFeatureCount(force=True)
    return feat_count


def _get_feature_counts(
    data_sources: Tuple[DataSource, ...],
    layer_names: Tuple[str, ...],
) -> Tuple[int, ...]:
    return tuple(
        _get_feature_count(data_source=data_source, layer_name=layer_name)
        for data_source, layer_name in zip(  # noqa: B905 no strict parameter in 3.8.10
            data_sources, layer_names
        )
    )


def _get_chunks(feature_count: int, ideal_chunk_size: int) -> Chunks:
    exclusive_range = range(0, feature_count, ideal_chunk_size)
    inclusive_range = tuple(exclusive_range) + (feature_count + 1,)
    range_pairs = pairwise(inclusive_range)
    return tuple(range_pairs)


def _get_sequence_of_chunks(
    feature_counts: Tuple[int, ...],
    ideal_chunk_size: int,
) -> Tuple[Chunks, ...]:
    return tuple(
        _get_chunks(feature_count=feature_count, ideal_chunk_size=ideal_chunk_size)
        for feature_count in feature_counts
    )


def _get_total_chunks(sequence_of_chunks: Tuple[Chunks, ...]) -> int:
    return sum(len(chunks) for chunks in sequence_of_chunks)


def _create_paths_sdf(
    spark: SparkSession,
    paths: Tuple[str, ...],
) -> SparkDataFrame:
    rows = tuple(Row(path=path) for path in paths)

    return spark.createDataFrame(
        data=rows,
        schema=StructType(
            [
                StructField("path", StringType()),
            ]
        ),
    )


def _create_chunks_sdf(
    spark: SparkSession,
    paths: Tuple[str, ...],
    layer_names: Tuple[str, ...],
    sequence_of_chunks: Tuple[Chunks, ...],
) -> SparkDataFrame:
    rows = tuple(
        Row(
            path=path,
            layer_name=layer_name,
            chunks=chunks,
        )
        for path, layer_name, chunks in zip(  # noqa: B905 no strict parameter in 3.8.10
            paths,
            layer_names,
            sequence_of_chunks,
        )
    )

    sdf = spark.createDataFrame(
        data=rows,
        schema=StructType(
            [
                StructField("path", StringType()),
                StructField("layer_name", StringType()),
                StructField("chunks", ArrayType(ArrayType(IntegerType()))),
            ]
        ),
    )

    return (
        sdf.withColumn("chunk", explode("chunks"))
        .withColumn("id", monotonically_increasing_id())
        .withColumn("start", col("chunk")[0])
        .withColumn("stop", col("chunk")[1])
        .drop("chunk", "chunks")
    )



"""Functions for reading a GeoPackage using Spark's JDBC drivers.

Reading a GeoPackage using Spark's JDBC drivers.
===================================================

Register the GeoPackage Dialect
-------------------------------

To read a GeoPackage using Spark's JDBC drivers, you will need to register a custom
mapping of GeoPackage to Spark Catalyst types.


To do this, run the following in a Databricks notebook cell:

.. code-block:: scala

    %scala
    import org.apache.spark.sql.jdbc.{JdbcDialect, JdbcDialects}
    import org.apache.spark.sql.types._

    object GeoPackageDialect extends JdbcDialect {
    override def canHandle(url: String): Boolean = url.startsWith("jdbc:sqlite")

    override def getCatalystType(
        sqlType: Int,
        typeName: String,
        size: Int,
        md: MetadataBuilder
    ): Option[DataType] = typeName match {
        case "BOOLEAN" => Some(BooleanType)
        case "TINYINT" => Some(ByteType)
        case "SMALLINT" => Some(ShortType)
        case "MEDIUMINT" => Some(IntegerType)
        case "INT" | "INTEGER" => Some(LongType)
        case "FLOAT" => Some(FloatType)
        case "DOUBLE" | "REAL" => Some(DoubleType)
        case "TEXT" => Some(StringType)
        case "BLOB" => Some(BinaryType)
        case "GEOMETRY" | "POINT" | "LINESTRING" | "POLYGON" | "MULTIPOINT" |
        "MULTILINESTRING" | "MULTIPOLYGON" | "GEOMETRYCOLLECTION" | "CIRCULARSTRING" |
        "COMPOUNDCURVE" | "CURVEPOLYGON" | "MULTICURVE" | "MULTISURFACE" | "CURVE" |
        "SURFACE" => Some(BinaryType)
        case "DATE" => Some(DateType)
        case "DATETIME" => Some(StringType)
    }
    }

    JdbcDialects.registerDialect(GeoPackageDialect)

Once you've registered the GeoPackage dialect, you can list the layers in a given
GeoPackage, get the coordinate reference system of a given layer, or read a layer
into a Spark DataFrame with the geometry column encoded as Well Known Binary.

List layers
-----------

To list the layers in a given GeoPackage, use :func:`list_layers`:

.. code-block:: python

    list_layers(
        path="/path/to/file.gpkg"
    )

Get the coordinate reference system of a layer
----------------------------------------------

To get the coordinate reference system of a layers in a given GeoPackage, use
:func:`get_crs`:

.. code-block:: python

    get_crs(
        path="/path/to/file.gpkg"
        layer_name="layer_name"
    )

Get the bounding box of a layer
-------------------------------

To get the bounding box of a layers in a given GeoPackage, use
:func:`get_bounds`:

.. code-block:: python

    get_bounds(
        path="/path/to/file.gpkg"
        layer_name="layer_name"
    )

Read a layer into a Spark DataFrame
-----------------------------------

To read a layer into a Spark DataFrame, use :func:`read_gpkg`, If you don't supply
a layer, the first layer will be read:

.. code-block:: python

    read_gpkg(
        path="/path/to/file.gpkg"
    )

Read a specific layer
^^^^^^^^^^^^^^^^^^^^^

If you supply a layer, that will be read:

.. code-block:: python

    read_gpkg(
        path="/path/to/file.gpkg"
        layer_name="layer_name"
   )

Supply a custom geometry column name
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The function will assume that the geometry column is called `geom`, if it's called
something else, supply a different `original_geometry_column_name`:

.. code-block:: python

    read_gpkg(
        path="/path/to/file.gpkg"
        layer_name="layer_name"
        original_geometry_column_name="geometry",
    )

GeoPackage Geometry Encoding
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

`GeoPackage Geometry Encoding`_ consists of two parts:

1. A GeoPackage Binary (GPB) header
2. A `Well Known Binary (WKB)`_ geometry

By default, the function will split the original geometry column into these two
parts, unpack the GPB header, and return both as, respectively, `gpb_header` and
`wkb_geometry`.

If you don't want the GPB header, set `drop_gpb_header` to `True`:

.. code-block:: python

    read_gpkg(
        path="/path/to/file.gpkg"
        layer_name="layer_name"
        drop_gpb_header=True,
    )

The length of the GPB header depends on the format of envelope (bounding box) included
in it. The function assumes that the envelope will be [minx, maxx, miny, maxy] and
that, therefore, the GPB header will be 40 bytes long. However, it's possible for the
format of the envelope to be [minx, maxx, miny, maxy, minz, maxz] or [minx, maxx, miny,
maxy, minm, maxm] and, therefore 56 bytes long, or [minx, maxx, miny, maxy, minz, maxz,
minm, maxm] and, therefore, 72 bytes long.

In these cases you will need to set `header_length` to the appropriate value:

.. code-block:: python

    read_gpkg(
        path="/path/to/file.gpkg"
        layer_name="layer_name"
        header_length=56,
    )

.. _`GeoPackage Geometry Encoding`:
    https://www.geopackage.org/spec/#gpb_format

.. _`Well Known Binary (WKB)`:
    https://libgeos.org/specifications/wkb/

"""
from functools import partial
from struct import unpack
from typing import Any, Dict, List, Optional, Tuple

from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql import SparkSession
from pyspark.sql.functions import array, col, collect_list, expr, udf
from pyspark.sql.types import (
    ArrayType,
    DoubleType,
    IntegerType,
    StringType,
    StructField,
    StructType,
)

HEADER_LENGTH = 40


def _read_gpkg(
    path: str,
    layer_name: str,
    spark: SparkSession,
) -> SparkDataFrame:
    """Read a given table from a GeoPackage/Sqlite file."""
    return (
        spark.read.format("jdbc")
        .option("url", f"jdbc:sqlite:{path}")
        .option("dbtable", layer_name)
        .load()
    )


_get_gpkg_contents = partial(
    _read_gpkg,
    layer_name="gpkg_contents",
)


return_schema = StructType(
    [
        StructField("magic", StringType()),
        StructField("version", IntegerType()),
        StructField("flags", StringType()),
        StructField("srs_id", IntegerType()),
        StructField("envelope", ArrayType(DoubleType())),
    ]
)


@udf(returnType=return_schema)
def _unpack_gpb_header(byte_array: bytearray) -> Dict[str, Any]:
    """Unpack the GeoPackage Binary Header."""
    unpacked = unpack("ccBBidddd", byte_array)
    return {
        "magic": unpacked[0].decode("ascii") + unpacked[1].decode("ascii"),
        "version": unpacked[2],
        "flags": format(unpacked[3], "b").zfill(8),
        "srs_id": unpacked[4],
        "envelope": [unpacked[5], unpacked[6], unpacked[7], unpacked[8]],
    }


def _split_geometry(
    df: SparkDataFrame,
    original_geometry_column_name: str,
    header_length: int,
    drop_gpb_header: bool,
) -> SparkDataFrame:
    """Split the GeoPackage geometry column into the binary header and WKB geometry."""
    split_df = df.select(
        "*",
        expr(
            f"SUBSTRING({original_geometry_column_name}, 1, {header_length}) AS gpb_header"  # noqa B950
        ),
        expr(
            f"SUBSTRING({original_geometry_column_name}, {header_length}+1, LENGTH({original_geometry_column_name})-{header_length}) AS wkb_geometry"  # noqa B950
        ),
    )
    if drop_gpb_header:
        return split_df.drop(original_geometry_column_name, "gpb_header")
    else:
        return split_df.drop(original_geometry_column_name).withColumn(
            "gpb_header", _unpack_gpb_header("gpb_header")
        )


def list_layers(
    path: str,
    spark: Optional[SparkSession] = None,
) -> List[str]:
    """List layers from a GeoPackage.

    Examples:
        >>> list_layers(
            "/dbfs/mnt/base/unrestricted/source_rpa_spatial_data_mart/dataset_rpa_reference_parcels/format_GPKG_rpa_reference_parcels/LATEST_rpa_reference_parcels/reference_parcels.gpkg"
        )
        ['reference_parcels']

    Args:
        path (str): Path to the GeoPackage. Must use the File API Format, i.e.
            start with `/dbfs` instead of `dbfs:`.
        spark (Optional[SparkSession], optional): The SparkSession to use. If none
             is provided, the active session will be used. Defaults to None.

    Returns:
        List[str]: A list of layer names.
    """  # noqa: B950
    _spark = spark if spark else SparkSession.getActiveSession()
    gpkg_contents = _get_gpkg_contents(
        path=path,
        spark=_spark,
    )
    # ! Explicitly coerce `table_names` to `List[str]` to prevent `mypy`
    # ! `[no-any-return]`
    table_names: List[str] = gpkg_contents.select(collect_list("table_name")).first()[0]
    return table_names


def get_crs(
    path: str,
    layer_name: str,
    spark: Optional[SparkSession] = None,
) -> str:
    """Get CRS of a layer in a GeoPackage.

    Examples:
        >>> get_crs(
            "/dbfs/mnt/base/unrestricted/source_rpa_spatial_data_mart/dataset_rpa_reference_parcels/format_GPKG_rpa_reference_parcels/LATEST_rpa_reference_parcels/reference_parcels.gpkg",
            layer_name="reference_parcels",
        )
        'PROJCS["OSGB 1936 / British National Grid",GEOGCS["OSGB 1936",DATUM["OSGB_1936",SPHEROID["Airy 1830",6377563.396,299.3249646,AUTHORITY["EPSG","7001"]],AUTHORITY["EPSG","6277"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4277"]],PROJECTION["Transverse_Mercator"],PARAMETER["latitude_of_origin",49],PARAMETER["central_meridian",-2],PARAMETER["scale_factor",0.9996012717],PARAMETER["false_easting",400000],PARAMETER["false_northing",-100000],UNIT["metre",1,AUTHORITY["EPSG","9001"]],AXIS["Easting",EAST],AXIS["Northing",NORTH],AUTHORITY["EPSG","27700"]]'

    Args:
        path (str): Path to the GeoPackage. Must use the File API Format, i.e.
            start with `/dbfs` instead of `dbfs:`.
        layer_name (str): The name of the layer to read.
        spark (Optional[SparkSession], optional): The SparkSession to use. If none
             is provided, the active session will be used. Defaults to None.

    Returns:
        str: `Well-known Text`_ Representation of the Spatial Reference System.

    .. _`Well-known Text`:
        https://www.ogc.org/standards/wkt-crs
    """  # noqa: B950
    _spark = spark if spark else SparkSession.getActiveSession()
    gpkg_contents = _get_gpkg_contents(
        path=path,
        spark=_spark,
    )
    srs_id = (
        gpkg_contents.filter(col("table_name") == layer_name)
        .select("srs_id")
        .first()[0]
    )
    gpkg_spatial_ref_sys = _read_gpkg(
        path=path,
        layer_name="gpkg_spatial_ref_sys",
        spark=_spark,
    )
    # ! Explicitly coerce `definition` to `str` to prevent `mypy` `[no-any-return]`
    definition: str = (
        gpkg_spatial_ref_sys.filter(col("srs_id") == srs_id)
        .select("definition")
        .first()[0]
    )
    return definition


def get_bounds(
    path: str,
    layer_name: str,
    spark: Optional[SparkSession] = None,
) -> Tuple[float, ...]:
    """Get bounding box of a layer in a GeoPackage.

    Examples:
        >>> get_bounds(
            "/dbfs/mnt/base/unrestricted/source_rpa_spatial_data_mart/dataset_rpa_reference_parcels/format_GPKG_rpa_reference_parcels/LATEST_rpa_reference_parcels/reference_parcels.gpkg",
            layer_name="reference_parcels",
        )
        (85668.93009999998, 7079.01, 654494.0698999999, 657493.4378)

    Args:
        path (str): Path to the GeoPackage. Must use the File API Format, i.e.
            start with `/dbfs` instead of `dbfs:`.
        layer_name (str): The name of the layer to read.
        spark (Optional[SparkSession], optional): The SparkSession to use. If none
             is provided, the active session will be used. Defaults to None.

    Returns:
        Tuple[float, ...]: a tuple containing `minx`, `miny`, `maxx`, `maxy` values
            for the layer.
    """  # noqa: B950
    _spark = spark if spark else SparkSession.getActiveSession()
    gpkg_contents = _get_gpkg_contents(
        path=path,
        spark=_spark,
    )
    # ! Explicitly coerce `bounds` to `Tuple[float, ...]` to prevent `mypy`
    # ! `[no-any-return]`
    bounds: Tuple[float, ...] = tuple(
        gpkg_contents.filter(col("table_name") == layer_name)
        .select(array(col("min_x"), col("min_y"), col("max_x"), col("max_y")))
        .first()[0]
    )
    return bounds


def read_gpkg(
    path: str,
    original_geometry_column_name: str = "geom",
    header_length: int = HEADER_LENGTH,
    drop_gpb_header: bool = False,
    layer_name: Optional[str] = None,
    spark: Optional[SparkSession] = None,
) -> SparkDataFrame:
    """Read a GeoPackage into Spark DataFrame.

    Args:
        path (str): Path to the GeoPackage. Must use the File API Format, i.e.
            start with `/dbfs` instead of `dbfs:`.
        original_geometry_column_name (str): The name of the geometry column
            of the layer in the GeoPackage. Defaults to "geom".
        header_length (int): The length of the `GPB header`_, in bytes.
            Defaults to 40.
        drop_gpb_header (bool): Whether to exclude the `GPB header`_ from
            the returned DataFrame. Defaults to False.
        layer_name (str, optional): The name of the layer to read. If none is provided,
            the first layer is read. Defaults to None.
        spark (SparkSession, optional): The SparkSession to use. If none
             is provided, the active session will be used. Defaults to None.

    Returns:
        SparkDataFrame: A Spark DataFrame with geometry encoded as
            `Well Known Binary (WKB)`_.

    .. _`GPB header`:
        https://www.geopackage.org/spec/#gpb_format
    .. _`Well Known Binary (WKB)`:
        https://libgeos.org/specifications/wkb/
    """
    _spark = spark if spark else SparkSession.getActiveSession()

    _layer_name = (
        layer_name
        if layer_name
        else list_layers(
            path=path,
            spark=_spark,
        )[0]
    )

    df = _read_gpkg(
        path=path,
        layer_name=_layer_name,
        spark=_spark,
    )

    return _split_geometry(
        df=df,
        original_geometry_column_name=original_geometry_column_name,
        header_length=header_length,
        drop_gpb_header=drop_gpb_header,
    )



from types import MappingProxyType
from typing import Optional, Tuple, Union

from osgeo.ogr import DataSource, Layer, Open
from pyspark.sql.types import StructField, StructType

from pyspark_vector_files._files import _get_layer_name


def _get_layer(
    data_source: DataSource,
    layer_name: str,
    start: Optional[int],
    stop: Optional[int],
) -> Layer:
    """Returns a GDAL Layer from a SQL statement, name or index, or 0th layer."""
    # ! "S608: Possible SQL injection vector through string-based query
    # ! construction" is turned off here because `_get_layer_name` will
    # ! raise an error if the user supplied layer doesn't exist and
    # ! `start` and `stop` are generated within the function.
    if isinstance(start, int) and isinstance(stop, int):
        sql = f'SELECT * from "{layer_name}" WHERE FID >= {start} AND FID < {stop}'  # noqa: S608, B950
    else:
        sql = f'SELECT * from "{layer_name}"'  # noqa: S608

    return data_source.ExecuteSQL(sql)


def _get_property_names(layer: Layer) -> Tuple[str, ...]:
    """Given a GDAL Layer, return the non-geometry field names."""
    layer_definition = layer.GetLayerDefn()
    return tuple(
        layer_definition.GetFieldDefn(index).GetName()
        for index in range(layer_definition.GetFieldCount())
    )


def _get_property_types(layer: Layer) -> Tuple[Tuple[int, int], ...]:
    """Given a GDAL Layer, return the non-geometry field types."""
    layer_definition = layer.GetLayerDefn()
    field_definitions = tuple(
        layer_definition.GetFieldDefn(index)
        for index in range(layer_definition.GetFieldCount())
    )
    return tuple(
        (field_definition.GetType(), field_definition.GetSubType())
        for field_definition in field_definitions
    )


def _get_feature_schema(
    layer: Layer,
    ogr_to_spark_type_map: MappingProxyType,
    geom_field_name: str,
    geom_field_type: Tuple[int, int],
) -> StructType:
    """Given a GDAL Layer and a data type mapping, return a PySpark DataFrame schema."""
    property_names = _get_property_names(layer=layer)
    property_types = _get_property_types(layer=layer)
    property_struct_fields = [
        StructField(field_name, ogr_to_spark_type_map[field_type])
        for field_name, field_type in zip(  # noqa: B905 no strict parameter in 3.8.10
            property_names, property_types
        )
    ]
    geometry_struct_field = [
        StructField(geom_field_name, ogr_to_spark_type_map[geom_field_type])
    ]
    return StructType(property_struct_fields + geometry_struct_field)


def _create_schema_for_files(
    path: str,
    layer_identifier: Optional[Union[str, int]],
    geom_field_name: str,
    geom_field_type: Tuple[int, int],
    ogr_to_spark_type_map: MappingProxyType,
) -> StructType:
    """Returns a schema for a given layer in the first file in a list of file paths."""
    data_source = Open(path)

    layer_name = _get_layer_name(
        # ! first argument to singledispatch function must be positional
        layer_identifier,
        data_source=data_source,
    )

    layer = _get_layer(
        data_source=data_source,
        layer_name=layer_name,
        start=None,
        stop=None,
    )

    return _get_feature_schema(
        layer=layer,
        ogr_to_spark_type_map=ogr_to_spark_type_map,
        geom_field_name=geom_field_name,
        geom_field_type=geom_field_type,
    )


def _create_schema_for_chunks(
    data_source: Optional[DataSource],
    layer_name: str,
    geom_field_name: str,
    geom_field_type: Tuple[int, int],
    ogr_to_spark_type_map: MappingProxyType,
) -> StructType:
    """Returns a schema for a given layer in the first file in a list of file paths."""
    layer = _get_layer(
        data_source=data_source,
        layer_name=layer_name,
        start=None,
        stop=None,
    )

    return _get_feature_schema(
        layer=layer,
        ogr_to_spark_type_map=ogr_to_spark_type_map,
        geom_field_name=geom_field_name,
        geom_field_type=geom_field_type,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pyspark_vector_files']

package_data = \
{'': ['*']}

install_requires = \
['GDAL==3.4.3',
 'more-itertools>=8.12.0,<9.0.0',
 'pandas>=1.2.4,<2.0.0',
 'pyspark==3.2.1']

setup_kwargs = {
    'name': 'pyspark-vector-files',
    'version': '0.2.5',
    'description': 'Read vector files into a Spark DataFrame with geometry encoded as WKB.',
    'long_description': '[![CI](https://github.com/Defra-Data-Science-Centre-of-Excellence/pyspark-vector-files/actions/workflows/ci.yml/badge.svg)](https://github.com/Defra-Data-Science-Centre-of-Excellence/pyspark-vector-files/actions/workflows/ci.yml)\n\n# PySpark Vector Files\n\nRead [vector files](https://gdal.org/drivers/vector/index.html) into a [Spark DataFrame](https://spark.apache.org/docs/latest/sql-programming-guide.html#datasets-and-dataframes) with geometry encoded as [Well Known Binary (WKB)](https://libgeos.org/specifications/wkb/).\n\nFull documentation is available [here](https://defra-data-science-centre-of-excellence.github.io/pyspark-vector-files/).\n\n## Requirements\n\nThis library was developed using [Databricks Runtime 10.4 LTS](https://docs.databricks.com/release-notes/runtime/10.4.html) and uses the versions of `python`, `pandas` and `pyspark` that come pre-installed on that runtime. However, it also requires `GDAL 3.4.3` as this is the most recent version of `GDAL` available from [ubuntugis-unstable](https://launchpad.net/~ubuntugis/+archive/ubuntu/ubuntugis-unstable) as of 2022-08-11.\n\nYou can install `GDAL` on your cluster using an [init script](https://docs.microsoft.com/en-us/azure/databricks/clusters/init-scripts). See [here](config/install_gdal.sh) for an example.\n\n## Install `pyspark-vector-files`\n\n### Within a Databricks notebook\n\n```sh\n%pip install pyspark-vector-files\n```\n\n### From the command line\n\n```sh\npython -m pip install pyspark-vector-files\n```\n\n## Quick start\n\nRead the first layer from a file or files with given extension into a single Spark DataFrame:\n\n```python\nfrom pyspark_vector_files import read_vector_files\n\nsdf = read_vector_files(\n    path="/path/to/files/",\n    suffix=".ext",\n)\n```\n\nMore examples are available [here](https://defra-data-science-centre-of-excellence.github.io/pyspark-vector-files/usage.html).\n\n## Local development\n\nTo ensure compatibility with [Databricks Runtime 10.4 LTS](https://docs.databricks.com/release-notes/runtime/10.4.html), this package was developed on a Linux machine running the `Ubuntu 20.04 LTS` operating system using `Python3.8.10`, `GDAL 3.4.3`, and `spark 3.2.1.`.\n\n### Install `Python 3.8.10` using [pyenv](https://github.com/pyenv/pyenv)\n\nSee the `pyenv-installer`\'s [Installation / Update / Uninstallation](https://github.com/pyenv/pyenv-installer#installation--update--uninstallation) instructions.\n\nInstall Python 3.8.10 globally:\n\n```sh\npyenv install 3.8.10\n```\n\nThen install it locally in the repository you\'re using:\n\n```sh\npyenv local 3.8.10\n```\n\n### Install `GDAL 3.4.3`\n\nAdd the [UbuntuGIS unstable Private Package Archive (PPA)](https://launchpad.net/~ubuntugis/+archive/ubuntu/ubuntugis-unstable)\nand update your package list:\n\n```sh\nsudo add-apt-repository ppa:ubuntugis/ubuntugis-unstable \\\n    && sudo apt-get update\n```\n\nInstall `gdal 3.4.3`, I found I also had to install python3-gdal (even though\nI\'m going to use poetry to install it in a virtual environment later) to\navoid version conflicts:\n\n```sh\nsudo apt-get install -y gdal-bin=3.4.3+dfsg-1~focal0 \\\n    libgdal-dev=3.4.3+dfsg-1~focal0 \\\n    python3-gdal=3.4.3+dfsg-1~focal0\n```\n\nVerify the installation:\n\n```sh\nogrinfo --version\n# GDAL 3.4.3, released 2022/04/22\n```\n\n### Install `poetry 1.1.13`\n\nSee poetry\'s [osx / linux / bashonwindows install instructions](https://python-poetry.org/docs/#osx--linux--bashonwindows-install-instructions)\n\n### Clone this repository\n\n```sh\ngit clone https://github.com/Defra-Data-Science-Centre-of-Excellence/pyspark_vector_files.git\n```\n\n### Install dependencies using `poetry`\n\n```sh\npoetry install\n```\n',
    'author': 'Ed Fawcett-Taylor',
    'author_email': 'ed.fawcett-taylor@defra.gov.uk',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Defra-Data-Science-Centre-of-Excellence/pyspark-vector-files',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.9',
}


setup(**setup_kwargs)



from itertools import compress
from types import MappingProxyType
from typing import Callable, Generator, Optional, Tuple, Union

from osgeo.ogr import Feature, Layer, Open, wkbNDR
from pandas import DataFrame as PandasDataFrame
from pandas import Series
from pyspark.sql.types import DataType, StructType

from pyspark_vector_files._files import _get_layer_name
from pyspark_vector_files._schema import _get_layer, _get_property_names


def _null_data_frame_from_schema(
    schema: StructType,
    spark_to_pandas_type_map: MappingProxyType,
) -> PandasDataFrame:
    """Generates an empty DataFrame that fits the schema."""
    return PandasDataFrame(
        data={
            field.name: Series(dtype=spark_to_pandas_type_map[field.dataType])
            for field in schema.fields
        }
    )


def _get_properties(feature: Feature) -> Tuple:
    """Given a GDAL Feature, return the non-geometry fields."""
    return tuple(field for field in feature)


def _get_geometry(feature: Feature) -> Tuple[Optional[bytearray]]:
    """Given a GDAL Feature, return the geometry field, if there is one."""
    geometry = feature.GetGeometryRef()
    if geometry:
        return (geometry.ExportToWkb(wkbNDR),)
    else:
        return (None,)


def _get_features(layer: Layer) -> Generator:
    """Given a GDAL Layer, return the all fields."""
    return ((_get_properties(feature) + _get_geometry(feature)) for feature in layer)


def _get_field_details(schema: StructType) -> Tuple[Tuple[str, DataType], ...]:
    """Returns fields from schema."""
    return tuple((field.name, field.dataType) for field in schema.fields)


def _get_field_names(
    schema_field_details: Tuple[Tuple[str, DataType], ...],
) -> Tuple[str, ...]:
    """Returns field names from schema fields."""
    return tuple(field[0] for field in schema_field_details)


def _get_columns_names(
    pdf: PandasDataFrame,
) -> Tuple[str, ...]:
    """Returns columns names from DataFrame."""
    return tuple(column for column in pdf.columns)


def _identify_missing_columns(
    schema_field_names: Tuple[str, ...],
    column_names: Tuple[str, ...],
) -> Tuple[bool, ...]:
    """Returns boolean mask of schema field names not in column names."""
    return tuple(field_name not in column_names for field_name in schema_field_names)


def _identify_additional_columns(
    schema_field_names: Tuple[str, ...],
    column_names: Tuple[str, ...],
) -> Tuple[bool, ...]:
    """Returns boolean mask of column names not in schema field names."""
    return tuple(column not in schema_field_names for column in column_names)


def _drop_additional_columns(
    pdf: PandasDataFrame,
    column_names: Tuple,
    additional_columns: Tuple,
) -> PandasDataFrame:
    """Removes additional columns from pandas DataFrame."""
    # ! columns has to be a list
    to_drop = list(compress(column_names, additional_columns))
    return pdf.drop(columns=to_drop)


def _coerce_types_to_schema(
    pdf: PandasDataFrame,
    schema_field_details: Tuple,
    spark_to_pandas_type_map: MappingProxyType,
) -> PandasDataFrame:
    for column_name, data_type in schema_field_details:
        pdf[column_name] = pdf[column_name].astype(spark_to_pandas_type_map[data_type])
    return pdf


def _add_missing_columns(
    pdf: PandasDataFrame,
    schema_field_names: Tuple,
    missing_columns: Tuple,
    schema_field_details: Tuple,
    spark_to_pandas_type_map: MappingProxyType,
) -> PandasDataFrame:
    """Adds missing fields to pandas DataFrame."""
    to_add = compress(schema_field_details, missing_columns)
    missing_column_series = tuple(
        Series(
            name=field_name,
            dtype=spark_to_pandas_type_map[field_type],
        )
        for field_name, field_type in to_add
    )
    pdf_plus_missing_columns = pdf.append(missing_column_series)
    reindexed_pdf = pdf_plus_missing_columns.reindex(columns=schema_field_names)
    coerced_pdf = _coerce_types_to_schema(
        pdf=reindexed_pdf,
        schema_field_details=schema_field_details,
        spark_to_pandas_type_map=spark_to_pandas_type_map,
    )
    return coerced_pdf


def _coerce_columns_to_schema(
    pdf: PandasDataFrame,
    schema_field_details: Tuple,
    spark_to_pandas_type_map: MappingProxyType,
) -> PandasDataFrame:
    """Adds missing fields or removes additional columns to match schema."""
    schema_field_names = _get_field_names(
        schema_field_details=schema_field_details,
    )

    column_names = _get_columns_names(
        pdf=pdf,
    )

    if column_names == schema_field_names:
        return pdf

    missing_columns = _identify_missing_columns(
        schema_field_names=schema_field_names,
        column_names=column_names,
    )

    additional_columns = _identify_additional_columns(
        schema_field_names=schema_field_names,
        column_names=column_names,
    )

    if any(missing_columns) and any(additional_columns):
        pdf_minus_additional_columns = _drop_additional_columns(
            pdf=pdf,
            column_names=column_names,
            additional_columns=additional_columns,
        )
        return _add_missing_columns(
            pdf=pdf_minus_additional_columns,
            schema_field_details=schema_field_details,
            missing_columns=missing_columns,
            spark_to_pandas_type_map=spark_to_pandas_type_map,
            schema_field_names=schema_field_names,
        )

    elif any(missing_columns):
        return _add_missing_columns(
            pdf=pdf,
            schema_field_details=schema_field_details,
            missing_columns=missing_columns,
            spark_to_pandas_type_map=spark_to_pandas_type_map,
            schema_field_names=schema_field_names,
        )

    else:
        return _drop_additional_columns(
            pdf=pdf,
            column_names=column_names,
            additional_columns=additional_columns,
        )


def _coerce_pdf_to_schema(
    pdf: PandasDataFrame,
    schema: StructType,
    spark_to_pandas_type_map: MappingProxyType,
) -> PandasDataFrame:
    schema_field_details = _get_field_details(schema=schema)

    coerced_columns_pdf = _coerce_columns_to_schema(
        pdf=pdf,
        schema_field_details=schema_field_details,
        spark_to_pandas_type_map=spark_to_pandas_type_map,
    )

    return _coerce_types_to_schema(
        pdf=coerced_columns_pdf,
        schema_field_details=schema_field_details,
        spark_to_pandas_type_map=spark_to_pandas_type_map,
    )


def _pdf_from_layer(
    layer: Layer,
    geom_field_name: str,
) -> PandasDataFrame:
    features_generator = _get_features(layer=layer)
    feature_names = _get_property_names(layer=layer) + (geom_field_name,)
    return PandasDataFrame(data=features_generator, columns=feature_names)


def _pdf_from_vector_file(
    path: str,
    layer_identifier: Optional[Union[str, int]],
    geom_field_name: str,
    coerce_to_schema: bool,
    schema: StructType,
    spark_to_pandas_type_map: MappingProxyType,
) -> PandasDataFrame:
    """Given a file path and layer, returns a pandas DataFrame."""
    data_source = Open(path)

    if data_source is None:
        return _null_data_frame_from_schema(
            schema=schema,
            spark_to_pandas_type_map=spark_to_pandas_type_map,
        )

    try:
        layer_name = _get_layer_name(
            # ! first argument to singledispatch function must be positional
            layer_identifier,
            data_source=data_source,
        )
    except ValueError:
        return _null_data_frame_from_schema(
            schema=schema,
            spark_to_pandas_type_map=spark_to_pandas_type_map,
        )

    layer = _get_layer(
        data_source=data_source,
        layer_name=layer_name,
        start=None,
        stop=None,
    )

    pdf = _pdf_from_layer(
        layer=layer,
        geom_field_name=geom_field_name,
    )

    if coerce_to_schema:
        return _coerce_pdf_to_schema(
            pdf=pdf,
            schema=schema,
            spark_to_pandas_type_map=spark_to_pandas_type_map,
        )

    return pdf


def _pdf_from_vector_file_chunk(
    path: str,
    layer_name: str,
    start: int,
    stop: int,
    geom_field_name: str,
    coerce_to_schema: bool,
    schema: StructType,
    spark_to_pandas_type_map: MappingProxyType,
) -> PandasDataFrame:
    """Given a file path and layer, returns a pandas DataFrame."""
    data_source = Open(path)

    if data_source is None:
        return _null_data_frame_from_schema(
            schema=schema,
            spark_to_pandas_type_map=spark_to_pandas_type_map,
        )

    layer = _get_layer(
        data_source=data_source,
        layer_name=layer_name,
        start=start,
        stop=stop,
    )

    if layer is None:
        return _null_data_frame_from_schema(
            schema=schema,
            spark_to_pandas_type_map=spark_to_pandas_type_map,
        )

    pdf = _pdf_from_layer(
        layer=layer,
        geom_field_name=geom_field_name,
    )

    if coerce_to_schema:
        return _coerce_pdf_to_schema(
            pdf=pdf,
            schema=schema,
            spark_to_pandas_type_map=spark_to_pandas_type_map,
        )

    return pdf


def _generate_parallel_reader_for_files(
    layer_identifier: Optional[Union[str, int]],
    geom_field_name: str,
    coerce_to_schema: bool,
    schema: StructType,
    spark_to_pandas_type_map: MappingProxyType,
) -> Callable[[PandasDataFrame], PandasDataFrame]:
    """Adds arbitrary key word arguments to the wrapped function."""

    def _(pdf: PandasDataFrame) -> PandasDataFrame:
        """Returns a pandas_udf compatible version of _pdf_from_vector_file."""
        return _pdf_from_vector_file(
            path=str(pdf["path"][0]),
            layer_identifier=layer_identifier,
            geom_field_name=geom_field_name,
            coerce_to_schema=coerce_to_schema,
            schema=schema,
            spark_to_pandas_type_map=spark_to_pandas_type_map,
        )

    return _


def _generate_parallel_reader_for_chunks(
    geom_field_name: str,
    coerce_to_schema: bool,
    schema: StructType,
    spark_to_pandas_type_map: MappingProxyType,
) -> Callable[[PandasDataFrame], PandasDataFrame]:
    """Adds arbitrary key word arguments to the wrapped function."""

    def _(pdf: PandasDataFrame) -> PandasDataFrame:
        """Returns a pandas_udf compatible version of _pdf_from_vector_file_chunk."""
        return _pdf_from_vector_file_chunk(
            path=str(pdf["path"][0]),
            layer_name=str(pdf["layer_name"][0]),
            start=int(pdf["start"][0]),
            stop=int(pdf["stop"][0]),
            geom_field_name=geom_field_name,
            coerce_to_schema=coerce_to_schema,
            schema=schema,
            spark_to_pandas_type_map=spark_to_pandas_type_map,
        )

    return _


# def _set_geometry_column_metadata(
#     df: DataFrame,
#     geometry_column_name: str,
#     layer: Layer,
# ) -> None:
#     df.schema[geometry_column_name].metadata = {
#         "crs": layer.GetSpatialRef().ExportToWkt(),
#         "encoding": "WKT",
#         "bbox": layer.GetExtent(),
#     }



from enum import Enum
from typing import Tuple


class ConcurrencyStrategy(Enum):
    """Valid concurrency strategies."""

    FILES = "files"
    ROWS = "rows"


Chunk = Tuple[int, int]
Chunks = Tuple[Chunk, ...]
